package me.yuxiaoyao.config.loader;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;

import java.io.*;
import java.util.*;

/**
 * @author Kerry on 19/10/09
 */

public class ConfigLoader {

    private static final String PROFILE = "profile";
    private static final String ACTIVE = "active";

    public static final String ACTIVE_PROFILE = PROFILE + "." + ACTIVE;

    /**
     * 如果想修改,可以通过反射修改
     */
    public static String DEFAULT_FILE_CONFIG_FOLDER = "config";
    private static String DEFAULT_APPLICATION = "application";

    private static String START_PLACEHOLDER = "${";
    private static String END_PLACEHOLDER = "}";

    private enum MapperFactoryEnum {
        /**
         * yml
         */
        YML("yml", "com.fasterxml.jackson.dataformat.yaml.YAMLFactory"),
        /**
         * yaml
         */
        YAML("yaml", "com.fasterxml.jackson.dataformat.yaml.YAMLFactory"),
        /**
         * properties
         */
        PROPERTIES("properties", "com.fasterxml.jackson.dataformat.javaprop.JavaPropsMapper"),
        /**
         * default json
         */
        JSON("json", ObjectMapper.class.getCanonicalName());
        private String name;
        private Class<?> factory;

        MapperFactoryEnum(String name, String className) {
            this.name = name;
            this.factory = findClass(className);
        }

        private static Class<?> findClass(String classname) {
            try {
                return Class.forName(classname);
            } catch (ClassNotFoundException ignored) {
            }
            return null;
        }

        public String getName() {
            return name;
        }

        public Class<?> getFactory() {
            return factory;
        }

        public ObjectMapper getObjectMapper() {
            if (factory == null) {
                return null;
            }
            try {
                final Object o = factory.newInstance();
                ObjectMapper objectMapper;
                if (this == MapperFactoryEnum.YML || this == MapperFactoryEnum.YAML) {
                    objectMapper = new ObjectMapper((JsonFactory) o);
                } else {
                    objectMapper = (ObjectMapper) o;
                }
                objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
                objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
                return objectMapper;
            } catch (InstantiationException | IllegalAccessException e) {
                e.printStackTrace();
            }
            return null;
        }
    }

    public interface ConfigListener {
        /**
         * 回调Profile
         *
         * @param profile
         */
        void activeProfile(String profile);

    }


    /**
     * 返回数据流
     *
     * @param path
     * @return 结果可能为 null
     */
    private static InputStream getConfigFileAsStream(String path) {
        try {
            // from config path
            final File file = new File(DEFAULT_FILE_CONFIG_FOLDER + File.separator + path);
            if (file.exists()) {
                return new FileInputStream(file);
            }
        } catch (FileNotFoundException ignored) {
        }
        // from class path
        return Thread.currentThread().getContextClassLoader().getResourceAsStream(path);
    }

    public static <T> T parseConfig(Class<T> clazz) {
        return parseConfig(null, clazz, null);
    }

    public static <T> T parseConfig(String[] args, Class<T> clazz, ConfigListener configListener) {
        Map map = loaderConfig(args, configListener);
        return getObjectMapper().convertValue(map, clazz);
    }


    public static <T> T parseConfig(TypeReference<T> toValueTypeRef) {
        return parseConfig(null, toValueTypeRef, null);
    }

    public static <T> T parseConfig(String[] args, TypeReference<T> toValueTypeRef, ConfigListener configListener) {
        Map map = loaderConfig(args, configListener);
        return getObjectMapper().convertValue(map, toValueTypeRef);
    }

    private static ObjectMapper getObjectMapper() {
        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        return objectMapper;
    }

    /**
     * 加载默认配置文件
     *
     * @return
     */
    public static Map loaderConfig() {
        return loaderConfig(null, null);
    }


    /**
     * 加载指定 profile 的配置文件
     *
     * @param args
     * @return
     */
    public static Map loaderConfig(String[] args, ConfigListener configListener) {

        Map defaultProperties = selectConfigPosition(DEFAULT_APPLICATION);
        if (defaultProperties == null) {
            return null;
        }
        String profile = getProfileByArgs(args);
        if (profile == null || profile.length() == 0) {
            profile = getActiveProfile(defaultProperties);
        }
        if (configListener != null) {
            // callback
            configListener.activeProfile(profile);
        }
        Map profileMap = selectConfigPosition(DEFAULT_APPLICATION + "-" + profile);
        if (profileMap != null) {
            removeActiveProfile(profileMap);
            //noinspection unchecked
            defaultProperties.putAll(profileMap);
        }
        return propertiesHandler(defaultProperties);
    }


    private static Map selectConfigPosition(String baseFilename) {
        InputStream configFileAsStream = null;
        MapperFactoryEnum parseType = null;
        for (MapperFactoryEnum value : MapperFactoryEnum.values()) {
            if (value.getFactory() == null) {
                continue;
            }
            configFileAsStream = getConfigFileAsStream(baseFilename + "." + value.getName());
            if (configFileAsStream != null) {
                parseType = value;
                break;
            }
        }
        if (parseType == null) {
            return null;
        }
        ObjectMapper objectMapper = parseType.getObjectMapper();
        if (objectMapper == null) {
            return null;
        }
        try {
            return objectMapper.readValue(configFileAsStream, LinkedHashMap.class);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                configFileAsStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        return null;
    }

    private static String getActiveProfile(Map fileMap) {
        Object profile = fileMap.get(PROFILE);
        if (profile instanceof Map) {
            final Object o = ((Map) profile).get(ACTIVE);
            if (o != null) {
                return (String) o;
            }
        }
        return null;
    }

    private static void removeActiveProfile(Map map) {
        Object profile = map.get(PROFILE);
        if (profile instanceof Map) {
            ((Map) profile).remove(ACTIVE);
        }
    }

    /**
     * 属性处理,目前暂时不处理,替换符号等操作
     *
     * @param map
     * @return
     */
    private static Map propertiesHandler(Map map) {
        Map<Object, String> preHandler = new LinkedHashMap<>();
        // find placeholder value
        findPreResolveProperties(map, preHandler, null);
        // replace value
        Map<Object, Object> objectObjectMap = ConfigLoader.resolveProperties(map, preHandler);
        // replace to map
        replaceProperties(map, objectObjectMap);
        return map;
    }


    private static Object getProperties(Map map, String key) {
        final int i = key.indexOf(".");
        if (i == -1) {
            return map.get(key);
        }
        String currentKey = key.substring(0, i);
        Object o = map.get(currentKey);

        if (o instanceof Map) {
            return getProperties((Map) o, (key.substring(i + 1)));
        }
        return o;
    }

    @SuppressWarnings("unchecked")
    private static void setProperties(Map map, String key, Object value) {
        if (map == null) {
            return;
        }
        final int i = key.indexOf(".");
        if (i == -1) {
            map.put(key, value);
            return;
        }
        String currentKey = key.substring(0, i);
        Object o = map.get(currentKey);
        if (o instanceof Map) {
            setProperties((Map) o, (key.substring(i + 1)), value);
        } else {
            map.put(key, value);
        }
    }


    public static void replaceProperties(Map map, Map<Object, ?> handlerMap) {

        for (Object o : handlerMap.keySet()) {
            Object value = handlerMap.get(o);
            setProperties(map, o.toString(), value);

        }

    }


    /**
     * 把带的占位符的值替换为真实值
     *
     * @param map
     * @param preHandlerMap
     * @return
     */
    public static Map<Object, Object> resolveProperties(Map map, Map<Object, ?> preHandlerMap) {
        Map<Object, Object> processedMap = new LinkedHashMap<>(preHandlerMap.size());

        for (Object o : preHandlerMap.keySet()) {
            Object value = preHandlerMap.get(o);

            if (value instanceof String) {
                // source value
                String str = ((String) value).trim();
                str = str.substring(START_PLACEHOLDER.length(), str.length() - END_PLACEHOLDER.length());
                String[] defaultValue = str.split(":");
                Object properties;
                if (defaultValue.length > 1) {
                    properties = getProperties(map, defaultValue[0]);
                } else {
                    properties = getProperties(map, defaultValue[0]);
                }
                // set default value
                if (properties == null) {
                    if (defaultValue.length > 1) {
                        properties = defaultValue[1];
                    } else {
                        properties = value;
                    }
                }
                /// System.out.println("key = [" + value + "] - value = [" + properties + "]");
                processedMap.put(o, properties);
            }
        }
        return processedMap;
    }

    /**
     * 查找所有具有占位符号的值,目前不支持数组
     * 起始符 {@link ConfigLoader#START_PLACEHOLDER}
     * 结束符 {@link ConfigLoader#END_PLACEHOLDER}
     *
     * @param map
     * @param result 查找结果接收
     * @return
     */
    public static void findPreResolveProperties(Map map, Map<Object, String> result, String prefix) {

        for (Object key : map.keySet()) {
            Object value = map.get(key);
            if (value instanceof Map) {
                findPreResolveProperties((Map) value, result, prefix == null ? key.toString() : prefix + "." + key.toString());
            } else if (value instanceof String) {
                String str = ((String) value).trim();
                if (str.startsWith(START_PLACEHOLDER) && str.endsWith(END_PLACEHOLDER)) {
                    if (prefix == null) {
                        result.put(key.toString(), str);
                    } else {
                        result.put(prefix + "." + key.toString(), str);
                    }
                }
            }
        }
        //prefix = "";
    }


    private static String getProfileByArgs(String[] args) {
        if (args != null && args.length > 0) {
            for (String arg : args) {
                arg = arg.trim();
                if (arg.startsWith("--profile.active=")) {
                    return arg.substring("--profile.active=".length());
                }
            }
        }
        return null;
    }


}
